# tt_limpa.rb
#-----------------------------------------------------------------------------
# Version: 1.1.2
# Compatible: SketchUp 7.1 (PC)
#-----------------------------------------------------------------------------
#
# CHANGELOG
# 1.0.0 - 21.10.2009
#		 * Simple search for coplanar edges.
#
# 1.0.1 - 22.10.2009
#		 * Decreased the tolerance.
#		 * Added option to respect materials.
#
# 1.1.0 - 23.10.2009
#		 * Limpa entire model.
#		   > Erase all coplanar edges
#		   > Purge
#		   > Remove Edge Materials
#		   > Smooth Edges
#
# 1.1.1 - 23.10.2009
#		 * Added option to erase lonely edges.
#
# 1.1.2 - 23.10.2009
#		 * Added option to erase lonely edges.
#
#-----------------------------------------------------------------------------
#
# TO DO
# * Add Plugin Manager support.
#
#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'

#-----------------------------------------------------------------------------


module TT_Cleanup
	
	# This plugin iterates over the given entities and looks for edges that
	# are located between coplanar faces.
	# Vector3d.same_direction? and .parallel? will return true for some edges
	# even though erasing the edge would make both faces disappear.
	# It appears to be inconsistancy between the SU Ruby API and that of what
	# SU actually uses. Therefor this match must be done manually. This is the
	# best tolerance I've managed to find so far.
	#
	# Dot = Normal1 % Normal2
	# Some results that returned 0.9999999991 was safe to erase, some where not.
	# Did not seem to get any smaller fractions of 0.9999999991 - so we test for
	# precision higher than this value. Anything produsing a dot product equal or
	# less than this isn't coplanar.
	#COPLANAR_TOLERANCE = 0.9999999991
	COPLANAR_TOLERANCE = 0.99999999999
	
	# Not coplanar
	# 0.999999999898786
	# 0.999999999901467
	# 0.99999999992968
	# 0.99999999995352
	# 0.999999999959393
	# 0.999999999974808
	# 0.999999999978347
	# 0.9999999991
	# 0.99999999999    
	# 0.00000000000    
	# x----------------
	# 1.0 - 1.0e-12
	# 0.999999999999
	# 1.0 - 0.000000000001000
	# 0.999999999999
	#
	# Coplanar
	# 0.999999999917881
	# 0.999999999918562
	# 0.999999999925243
	# 0.999999999960825
	# 0.999999999984962
	# 0.999999999999148
	# 0.999999999999408
	# 0.9999999991
	# x----------------
	
	V_MAJOR = 0
	V_MINOR = 1
	
	@mode_strict = false
	@mode_ignore_materials = false
	@lonely_edges = true
	
	@tolerance = COPLANAR_TOLERANCE
	
	unless file_loaded?('tt_limpa.rb')
	# Add some menu items to access this
	plugins_menu = UI.menu('Plugins')
	probes_menu = plugins_menu.add_submenu('Limpa')
	probes_menu.add_item('Limpa Entire Model')			{ self.limpa! }
	#probes_menu.add_separator
	probes_menu.add_item('Erase Coplanar Edges')			{ self.erase_coplanar }
	probes_menu.add_item('Erase Non Solid Faces')			{ self.erase_faces }
	probes_menu.add_separator
	menu_strict		= probes_menu.add_item('Same Direction')	{ @mode_strict = !@mode_strict }
	menu_ignore_mat	= probes_menu.add_item('Ignore Materials')	{ @mode_ignore_materials = !@mode_ignore_materials }
	menu_lone_edges	= probes_menu.add_item('Erase Lonely Edges'){ @lonely_edges = !@lonely_edges }
	# Validate procs
	probes_menu.set_validation_proc(menu_strict)		{ (@mode_strict) ? MF_CHECKED : MF_UNCHECKED }
	probes_menu.set_validation_proc(menu_ignore_mat)	{ (@mode_ignore_materials) ? MF_CHECKED : MF_UNCHECKED }
	probes_menu.set_validation_proc(menu_lone_edges)	{ (@lonely_edges) ? MF_CHECKED : MF_UNCHECKED }
	probes_menu.add_separator
	menu_strict		= probes_menu.add_item('Set Tolerance (Debug)')	{ self.set_tolerance }
	end
	
	# Process the whole model
	def self.limpa!
		# Any version less than 7.1 will be warned about unpredictable results.
		msg = 'Sketchup prior to 7.1 has a bug which might lead to loss of geometry. Do you want to continue?'
		return if self.version_check(7, 1, msg) == 7 # No
		
		# Clean up the model
		# (?) Remove/Transfer C/G materials (ignore V-Ray objects)
		
		prompts = ['Erase Coplanar Edges: ', ' Ignore Reversed Faces: ', ' Ignore Materials: ', 'Remove Lonely Edges: ', 'Remove Edge Materials: ', 'Smooth Edges by Angle : ', 'Purge: ' , 'Erase Non Solid Faces']
		@limpa_defaults = ['Yes', 'Yes', 'No', 'Yes', 'Yes', '30', 'Yes', 'Yes'] unless defined?(@limpa_defaults)
		lists = ['Yes|No', 'Yes|No', 'Yes|No', 'Yes|No', 'Yes|No', '', 'Yes|No', 'Yes|No']
		input = UI.inputbox(prompts, @limpa_defaults, lists, 'Clean Up Model')
		return if input == false
		
		@limpa_defaults = input
		puts input
		
		erase_coplanar			= (input[0] == 'Yes') ? true : false
		mode_strict				= (input[1] == 'Yes') ? false : true
		mode_ignore_materials	= (input[2] == 'Yes') ? true : false
		remove_lonely_edge		= (input[3] == 'Yes') ? true : false
		remove_edge_materials	= (input[4] == 'Yes') ? true : false
		smooth_edges			= (input[5].match(/^\d+(\.\d+)?$/).nil?) ? nil : input[5].to_f	
		purge					= (input[6] == 'Yes') ? true : false
		erase_faces         	= (input[7] == 'Yes') ? true : false
		
		model = Sketchup.active_model
		
		model.start_operation('Limpa', true)
		
		stats = {}
		stats['Elapsed Time'] = Time.now	
		
		# --- Purging --- #
		if purge
			stats['Purged Components'] = model.definitions.length
			Sketchup.status_text = 'Purging Components...'
			model.definitions.purge_unused
			stats['Purged Components'] -= model.definitions.length
		end		
			
		# --- Erase Coplanar Edges --- #
		if erase_coplanar
			stats['Erased Edges'] = 0
			stats['Faces Reduced'] = model.number_faces if model.respond_to?(:number_faces)
			edges = self.get_coplanar_edges(model.entities, mode_strict, mode_ignore_materials, remove_lonely_edge)
			Sketchup.status_text = "Erasing #{edges.length} coplanar edges. Please wait..."
			# Model context
			model.entities.erase_entities(edges)
			stats['Erased Edges'] += edges.length
			# Definition context
			model.definitions.each { |d|
				next if d.image?
				edges = self.get_coplanar_edges(d.entities, mode_strict, mode_ignore_materials, remove_lonely_edge)
				Sketchup.status_text = "Definition: #{d.name} - Erasing #{edges.length} coplanar edges. Please wait..."
				stats['Erased Edges'] += edges.length
				d.entities.erase_entities(edges)
			}
			stats['Faces Reduced'] -= model.number_faces if model.respond_to?(:number_faces)
		end
		
		
		
				# --- Erase Non Solid Faces --- #
		if erase_faces
			Sketchup.active_model.selection.to_a.each{|f|next if not f.valid? or not f.class==Sketchup::Face; f.edges.each{|e|e.erase! if e.valid? and not e.faces[1]}}
		end
		
		
		# --- Process the rest --- #
		"Processing Model. Please wait..."
		model.entities.each { |e|
			self.process_entity(e, remove_edge_materials, smooth_edges)
		}
		model.definitions.each { |d|
			Sketchup.status_text = "Processing #{d.name}. Please wait..."
			d.entities.each { |e|
				self.process_entity(e, remove_edge_materials, smooth_edges)
			}
		}
		
		# --- Purging --- #
		if purge			
			stats['Purged Layers'] = model.layers.length
			Sketchup.status_text = 'Purging Layers...'
			model.layers.purge_unused
			stats['Purged Layers'] -= model.layers.length
			
			stats['Purged Materials'] = model.materials.length
			Sketchup.status_text = 'Purging Materials...'
			model.materials.purge_unused
			stats['Purged Materials'] -= model.materials.length
			
			stats['Purged Styles'] = model.styles.count
			Sketchup.status_text = 'Purging Styles...'
			model.styles.purge_unused
			stats['Purged Styles'] -= model.styles.count
		end
		
		stats['Elapsed Time'] = Time.now - stats['Elapsed Time']
		Sketchup.status_text = 'Done!'
		model.commit_operation 
		
		# Stats
		UI.messagebox("Limpa Statistics:\n#{stats.collect{|k,v|"> #{k}: #{v}"}.sort.join("\n")}", MB_MULTILINE)
	end
	
	def self.process_entity(e, rem_mat=false, smooth=nil)
		if e.is_a?(Sketchup::Edge)
			# Remove Edge Material
			e.material = nil if rem_mat
			# Smooth Edge
			if !smooth.nil? && e.faces.length == 2
				angle = e.faces[0].normal.angle_between(e.faces[1].normal)
				if angle.radians.abs <= smooth
					e.smooth = true
					e.soft = true
				end
			end
		end
	end
	
	# Find all coplanar edges in the selection and select only them
	def self.select_coplanar
		sel = Sketchup.active_model.selection
		
		# If nothing is selected - pick the current context
		ents = (sel.length > 0) ? sel : Sketchup.active_model.active_entities
		
		edges = self.get_coplanar_edges(ents, @mode_strict, @mode_ignore_materials, @lonely_edges)
		sel.clear
		sel.add(edges)
	end
	
	
	
	
	
    # Find non solid faces in the selection and erase them
	
    def self.erase_faces
			Sketchup.active_model.selection.to_a.each{|f|next if not f.valid? or not f.class==Sketchup::Face; f.edges.each{|e|e.erase! if e.valid? and not e.faces[1]}}
	end
	# Find all coplanar edges in the selection and erase them
	def self.erase_coplanar
		# Any version less than 7.1 will be warned about unpredictable results.
		msg = 'Sketchup prior to 7.1 has a bug which might lead to loss of geometry. Do you want to continue?'
		return if self.version_check(7, 1, msg) == 7 # No
		
		model = Sketchup.active_model
		sel = model.selection
		
		# If nothing is selected - pick the current context
		ents = (sel.length > 0) ? sel : Sketchup.active_model.active_entities
		
		edges = self.get_coplanar_edges(ents, @mode_strict, @mode_ignore_materials, @lonely_edges)
		
		# Erase all coplanar edges.
		Sketchup.status_text = "Erasing #{edges.length} coplanar edges. Please wait..."
		model.entities.erase_entities(edges)
		Sketchup.status_text = "Erased #{edges.length} coplanar edges! Complete!"
	end
	
	# Find all coplanar edges in the given set of entities.
	# strict mode ensures the faces are in the same direction, otherwise it just checks for
	#		parallel faces. (front vs back -side is ignored.)
	# Also includes lonely faces not attached to any face.
	def self.get_coplanar_edges(entities, strict=false, ignore_materials=false, lonely_edges=true)
		Sketchup.status_text = 'Hunting for coplanar edges. Please wait...'
		c=[]
		entities.each { |e|
			next unless e.is_a?(Sketchup::Edge)
			# Remove lonely Edges.
			c << e if lonely_edges && e.faces.length == 0
			# Coplanar edges only have two faces connected.
			next unless e.faces.length == 2
			f1 = e.faces[0]
			f2 = e.faces[1]
			# We might want to consider the materials applied before removing any edge.
			unless ignore_materials
				next unless f1.material == f2.material && f1.back_material == f2.back_material
			end
			# "strict" means that we ensure the faces are point in the same direction. Otherwise
			# we ignore front/back side orientation.
			if strict
				#c << e if e.faces.length == 2 && e.faces[0].normal.samedirection?(e.faces[1].normal)
				c << e if f1.normal % f2.normal > @tolerance
			else
				#c << e if e.faces.length == 2 && e.faces[0].normal.parallel?(e.faces[1].normal)
				c << e if (f1.normal % f2.normal).abs > @tolerance
			end
		}
		return c
	end
	
	# Version Check
	def self.version_check(major, minor, message)
		v = Sketchup.version.split('.').map! { |i| i.to_i }
		return if v[V_MAJOR] > major
		return if v[V_MAJOR] == major && v[V_MINOR] >= minor
		return UI.messagebox(message, MB_YESNO)
	end
	
	# Debug UI
	def self.set_tolerance
		prompts = ['Tolerance']
		defaults = [@tolerance]
		input = UI.inputbox(prompts, defaults, 'Set new tolerance')
		@tolerance = input[0].to_f unless input == false
	end
end


#-----------------------------------------------------------------------------
file_loaded('tt_limpa.rb')
#-----------------------------------------------------------------------------